#!/usr/bin/python
# -*- coding: utf-8 -*-
#Python 3

## This file is part of Thor Battery Monitor (TBM).
## Open source monitoring of battery bank.
##
## Copyright (C) 2015-2018  TEROC AB.
##
## TBM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## TBM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with TBM.  If not, see <http://www.gnu.org/licenses/>.
##
## Note that TBM depends on other software, with different
## licensing conditions.
##
## We are thankful to Yoctopuce and others, who have provided helpful
## examples and support.
##
## For more information, contact Sven Ruin, sven.ruin@teroc.se

import os
import sys, configparser, logging, re
from http import client
## add ../../Sources to the PYTHONPATH
## sys.path.append(os.path.join("..","..","/home/pi/YoctoLib.python/Sources"))
from yoctopuce.yocto_api import *
from yoctopuce.yocto_voltage import *
from yoctopuce.yocto_display import *

print("Thor Battery Monitor \nVersion 1.00 \n")

## Runtime optimization global variables.
osnormpath = os.path.normpath
syspath = sys.path
logging.basicConfig(filename=osnormpath(syspath[0]+'/TBM.log'), filemode='a', level=logging.INFO)
config = configparser.ConfigParser()
timetime = time.time
loggingerror = logging.error
configwrite = config.write
research = re.search
YAPIUpdateDeviceList = YAPI.UpdateDeviceList
YAPISleep = YAPI.Sleep
YDisplayFirstDisplay = YDisplay.FirstDisplay
YSensorFindSensor = YSensor.FindSensor
YSensorFirstSensor = YSensor.FirstSensor
freqformat = "{0}/{1}".format
ecformat = "ec: {0}".format
commajoin = ",".join
colonformat = "{0}:{1}".format
curaverageaformat = "current_average_a: {0}".format
nettocycleformat = "netto_ah: {0}\ncycle_ah_out: {1}\ncycle_ah_in: {2}".format
percformat = "{0}%".format
modulecondisconformat = 'Module {0}connected: {1} ({2})'.format
sendArr = []
sendArrappend = sendArr.append

## The following settings are default (fallback) values.
## Change them as necessary by editing the TBM_settings.ini file created by
## this program.
##
## Domain you want to post to:
## localhost would be an emoncms installation on the same computer
## This could be emoncms.org to post to emoncms.org
domain = 'localhost'
##
## Location of emoncms in your server:
## the standard setup is to place it in a folder called emoncms
## To post to emoncms.org this would be blank: ''
emoncmspath = 'emoncms'
##
## Write API key of emoncms account
apikey = ''
##
## Node id you'd like the client to appear as
nodeid = 4
##
charged_v = 26.4
charged_a_ratio = 0.005
fully_charged_sec = 240.0
ah_cap_20 = 72.0
peukert_exp = 1.25
efficiency = 0.9
ah_in = 0.0
ah_out = 0.0
eff_calc_ah_ratio = 0.1
current_threshold_a = 0.0
loop_time_ms = 30000
save_to_db_time_sec = 1
low_bat_warn = 0.5
main_voltage_volt_name = 'UDC'
over_voltage = 29.0
under_voltage = 22.0
main_shunt_millivolt_name = 'IBB'
main_shunt_multiplier = 1.0
main_shunt_offset = 0.0
## re = Regular Expression (used to identify yoctopuce module functions)
## The functions are renamed via YoctoHub.
block_voltage_re = 'UB\d+'
cell_voltage_re = 'UC\d+'
block_deviation_beacon = 0.3
cell_deviation_beacon = 0.1

## Defining config functions.
def create_config():
    print("Creating TBM_settings.ini")
    config['EMONCMS'] = {'domain': domain,
                         'emoncmspath': emoncmspath,
                         'apikey': apikey,
                         'nodeid': str(nodeid)
    }
    config['SAVED VALUES'] = {
        'ah_cap_20': str(ah_cap_20),
        'ah_in': str(ah_in),
        'ah_out': str(ah_out),
        'block_deviation_beacon': str(block_deviation_beacon),
        'block_voltage_re': str(block_voltage_re),
        'cell_deviation_beacon': str(cell_deviation_beacon),
        'cell_voltage_re': str(cell_voltage_re),
        'charged_a_ratio': str(charged_a_ratio),
        'charged_v': str(charged_v),
        'current_threshold_a': str(current_threshold_a),
        'eff_calc_ah_ratio': str(eff_calc_ah_ratio),
        'efficiency': str(efficiency),
        'fully_charged_sec': str(fully_charged_sec),
        'loop_time_ms': str(loop_time_ms),
        'low_bat_warn': str(low_bat_warn),
        'main_shunt_millivolt_name': main_shunt_millivolt_name,
        'main_shunt_multiplier': str(main_shunt_multiplier),
        'main_shunt_offset': str(main_shunt_offset),
        'main_voltage_volt_name': main_voltage_volt_name,
        'over_voltage': str(over_voltage),
        'peukert_exp': str(peukert_exp),
        'save_to_db_time_sec': str(save_to_db_time_sec),
        'under_voltage': str(under_voltage)
    }
    with open(osnormpath(syspath[0]+'/TBM_settings.ini'), 'w') as configfile:
        configwrite(configfile)
#
def save_to_config():
    config['SAVED VALUES']['ah_in'] = str(ah_in)
    config['SAVED VALUES']['ah_out'] = str(ah_out)
    config['SAVED VALUES']['efficiency'] = str(efficiency)

    with open(osnormpath(syspath[0]+'/TBM_settings.ini'), 'w') as configfile:
        configwrite(configfile)
#
## Creating settings file with default values if it doesn't yet exist.
## If the settings file exists, read values to config var.
try:
    config.read_file(open(osnormpath(syspath[0]+"/TBM_settings.ini")))
## except FileNotFoundError:
    ## create_config()
except IOError:
    create_config()
except KeyError:
    create_config()
except Error:
    print("Unknown Error Reading Settings File")
    create_config()
#
configvars = config['SAVED VALUES']

## Get ini variables from config var.
##
## Single DB request version:
domain = config.get('EMONCMS','domain',fallback=domain)
emoncmspath = config.get('EMONCMS','emoncmspath',fallback=emoncmspath)
apikey = config.get('EMONCMS','apikey',fallback=apikey)
nodeid = config.getint('EMONCMS','nodeid',fallback=nodeid)
##
conn = client.HTTPConnection(domain)
##
connrequest = conn.request
connclose = conn.close
conngetresponse = conn.getresponse
##
## Multi DB requests version:
## conns = []
## domainpaths = [
             ## ['localhost','emoncms','APIKEY'],
             ## ['emoncms.org','','APIKEY'],
             ## ['172.16.3.99','emoncms','APIKEY']
           ## ]
## if (domainpaths):
    ## for pair in domainpaths:
        ## conns.append(client.HTTPConnection(pair[0]))
##
##
charged_v = configvars.getfloat('charged_v',
    fallback=charged_v)
charged_a_ratio = configvars.getfloat('charged_a_ratio',
    fallback=charged_a_ratio)
fully_charged_sec = configvars.getfloat('fully_charged_sec',
    fallback=fully_charged_sec)
ah_cap_20 = configvars.getfloat('ah_cap_20',
    fallback=ah_cap_20)
peukert_exp = configvars.getfloat('peukert_exp',
    fallback=peukert_exp)
efficiency = configvars.getfloat('efficiency',
    fallback=efficiency)
ah_in = configvars.getfloat('ah_in',
    fallback=ah_in)
ah_out = configvars.getfloat('ah_out',
    fallback=ah_out)
eff_calc_ah_ratio = configvars.getfloat('eff_calc_ah_ratio',
    fallback=eff_calc_ah_ratio)
current_threshold_a = configvars.getfloat('current_threshold_a',
    fallback=current_threshold_a)
loop_time_ms = configvars.getint('loop_time_ms',
    fallback=loop_time_ms)
save_to_db_time_sec = configvars.getfloat('save_to_db_time_sec',
    fallback=save_to_db_time_sec)
low_bat_warn = configvars.getfloat('low_bat_warn',
    fallback=low_bat_warn)
main_voltage_volt_name = configvars.get('main_voltage_volt_name',
    fallback=main_voltage_volt_name)
over_voltage = configvars.getfloat('over_voltage',
    fallback=over_voltage)
under_voltage = configvars.getfloat('under_voltage',
    fallback=under_voltage)
main_shunt_millivolt_name = configvars.get('main_shunt_millivolt_name',
    fallback=main_shunt_millivolt_name)
main_shunt_multiplier = configvars.getfloat('main_shunt_multiplier',
    fallback=main_shunt_multiplier)
main_shunt_offset = configvars.getfloat('main_shunt_offset',
    fallback=main_shunt_offset)
block_voltage_re = configvars.get('block_voltage_re',
    fallback=block_voltage_re)
cell_voltage_re = configvars.get('cell_voltage_re',
    fallback=cell_voltage_re)
block_deviation_beacon = configvars.getfloat('block_deviation_beacon',
    fallback=block_deviation_beacon)
cell_deviation_beacon = configvars.getfloat('cell_deviation_beacon',
    fallback=cell_deviation_beacon)
#
## Checking for unreasonable variable values, and listing errors if so.
varErrorList = []

if (charged_v <= 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini charged_v"
    )
#
if (charged_a_ratio <= 0.0 or charged_a_ratio > 0.1):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini charged_a_ratio"
    )
#
if (fully_charged_sec < (loop_time_ms/500)):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini fully_charged_sec"
    )
#
if (ah_cap_20 <= 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini ah_cap_20"
    )
#
if (peukert_exp < 1.0 or peukert_exp > 2.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini peukert_exp"
    )
#
if (efficiency > 1.0 or efficiency <= 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini efficiency"
    )
#
if (ah_in < 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini ah_in"
    )
#
if (ah_out < 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini ah_out"
    )
#
if (eff_calc_ah_ratio <= 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini eff_calc_ah_ratio"
    )
#
if (current_threshold_a < 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini current_threshold_a"
    )
#
if (loop_time_ms < 250 or loop_time_ms > 3600000):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini loop_time_ms"
    )
#
if (low_bat_warn < 0.0 or low_bat_warn >= 1.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini low_bat_warn"
    )
#
if (under_voltage < 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini under_voltage"
    )
#
if (over_voltage < under_voltage):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini over_voltage"
    )
#
if (main_shunt_multiplier <= 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini main_shunt_multiplier"
    )
#
if (block_deviation_beacon <= 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini block_deviation_beacon"
    )
#
if (cell_deviation_beacon <= 0.0):
    varErrorList.append(
       "Wrong parameter settings in TBM_settings.ini cell_deviation_beacon"
    )
#
if (len(varErrorList) > 0):
    sys.exit("\n".join(varErrorList))
#

errmsg = YRefParam()
## Setup the API to use the local USB devices
## Get access to your device, connected locally on USB for instance
if (YAPI.RegisterHub("usb",errmsg) != YAPI.SUCCESS):
    print(str(errmsg))
    if (YAPI.RegisterHub("127.0.0.1",errmsg) != YAPI.SUCCESS):
        sys.exit("init error: "+str(errmsg))
#
## Creating empty global lists.
modules = []
mod_millivolts = []
## mod_minidisplays = []
## mod_temps = []
mod_volts = []
sensor_arr = []
display_arr = []
block_voltage_arr = []
cell_voltage_arr = []

## Creating function variables to decrease execution time.
block_voltage_arrappend = block_voltage_arr.append
cell_voltage_arrappend = cell_voltage_arr.append
display_arrappend = display_arr.append
modulesappend = modules.append
modulesremove = modules.remove
mod_millivoltsappend = mod_millivolts.append
mod_voltsappend = mod_volts.append
sensor_arrappend = sensor_arr.append

## Creating some global variables
main_voltage_volt = None
main_shunt_millivolt = None
soc = -1.0 ## note below
## Since 0% is a valid SOC value, negative value will be used as false for 
## conditional expressions.
b_full_once = False
b_full_now = False
cycle_ah_in = 0.0
cycle_ah_out = 0.0
netto_ah = 0.0
calc_efficiency = 0.0 ## note below
## calc_efficiency is only used to calculate and print efficiency.
## 0.0 will be evaluated as a boolean false later on and
## therefor will not be printed until it has been calculated.

## Initializing time variables with current time.
starttime = timetime()
t_charged = starttime
t_send_to_db = starttime
last_report_time = 0.0
## And last report time to 0 to trigger device registration

## The following is old code that's been replaced.
##
## VoltSArr = []
## if (mod_volts):
     ## print("Connected Yocto-Volts:")
     ## for module in mod_volts:
         ## if (module.isOnline()):
             ## serialNumber = module.get_serialNumber()
             ## VoltSArr.append(YVoltage.FindVoltage(serialNumber + ".voltage1"))
             ## VoltSArr.append(YVoltage.FindVoltage(serialNumber + ".voltage2"))
             ## print(serialNumber)
##
## display_arr = []
## if (mod_minidisplays):
     ## print("Connected Yocto-MiniDisplays:")
     ## for module in mod_minidisplays:
         ## if (module.isOnline()):
             ## serialNumber = module.get_serialNumber()
             ## display_arr.append(YDisplay.FindDisplay(serialNumber + ".display"))
             ## print(serialNumber)
##
## if (mod_millivolts):
     ## print("Yocto-milliVolts:")
     ## for module in mod_millivolts:
         ## if (module.isOnline()):
             ## print(module.get_serialNumber())
##
## if (mod_temps):
     ## print("Yocto-Temperatures:")

     ## for module in mod_temps:
         ## if (module.isOnline()):
             ## print(module.get_serialNumber())
#

## Defining functions
def handle_error(e):
    print(type(e))
    print(e.args)
    print(e)
    loggingerror(type(e))
    loggingerror(e.args)
    loggingerror(e)
#
def low_bat_warn_func():
    ## Things to do when SOC is considered too low
    print("LOW BATTERY WARNING")
#
def over_voltage_warn_func():
    ## Things to do when voltage is too high
    print("OVERVOLTAGE WARNING")
#
def under_voltage_warn_func():
    ## Things to do when voltage is too low
    print("UNDERVOLTAGE WARNING")
#
def reg_modules():
    global block_voltage_arr, cell_voltage_arr, display_arr, sensor_arr, \
           main_shunt_millivolt, main_voltage_volt, mod_millivolts, mod_volts
    ## Clearing current module lists.
    ## modules.clear()
    ## mod_millivolts.clear()
    ## mod_volts.clear()
    ## sensor_arr.clear()
    ## display_arr.clear()
    ## block_voltage_arr.clear()
    ## cell_voltage_arr.clear()
    ## del method is faster than .clear()
    ## del modules[:]
    del mod_millivolts[:]
    del mod_volts[:]
    del sensor_arr[:]
    del display_arr[:]
    del block_voltage_arr[:]
    del cell_voltage_arr[:]
    main_voltage_volt = None
    main_shunt_millivolt = None
    
    ## module = YModule.FirstModule()
    ## while module is not None:
    if (modules):
        for module in modules:
            try:
                modName = module.get_productName()
                if (modName == "Yocto-milliVolt-Rx" 
                  or modName == "Yocto-milliVolt-Rx-BNC"):
                    mod_millivoltsappend(module)
                ## elif (modName == "Yocto-MiniDisplay"):
                    ## mod_minidisplays.append(module)
                ## elif (modName == "Yocto-Temperature"):
                    ## mod_temps.append(module)
                elif (modName == "Yocto-Volt"):
                    mod_voltsappend(module)
                module = module.nextModule()
            except Exception as exc:
                handle_error(exc)

    ## if (modules):
        ## print("\nConnected modules:")
        ## for module in modules:
            ## print(module.get_serialNumber()+' ('+module.get_productName()+')')

    this_sensor = YSensorFirstSensor()
    while this_sensor is not None:
        try:
            sensor_arrappend(this_sensor)
            s = this_sensor.get_logicalName()
            prodName = this_sensor.get_module().get_productName()
            if (research(block_voltage_re,s)):
                if (not block_voltage_arr or 
                  block_voltage_arr[0].get_module().get_productName() == prodName):
                    block_voltage_arrappend(this_sensor)
            elif (research(cell_voltage_re,s)):
                if (not cell_voltage_arr or 
                  cell_voltage_arr[0].get_module().get_productName() == prodName):
                    cell_voltage_arrappend(this_sensor)
            if (this_sensor.get_functionId() == "voltage1"):
                i = len(mod_volts)
                if (i == 1 or main_voltage_volt_name in s):
                    main_voltage_volt = this_sensor
            elif (prodName == "Yocto-milliVolt-Rx" 
               or prodName == "Yocto-milliVolt-Rx-BNC"):
                i = len(mod_millivolts)
                if (i == 1 or main_shunt_millivolt_name in s):
                    main_shunt_millivolt = this_sensor
                    freqStr = "2/m" ## example of 2 times a minute
                    if (loop_time_ms < 1000):
                        freqStr = freqformat(1000//loop_time_ms, "s")
                    elif (loop_time_ms < 60000):
                        freqStr = freqformat(60000//loop_time_ms, "m")
                    elif (loop_time_ms < 3600000):
                        freqStr = freqformat(3600000//loop_time_ms, "h")
                    else:
                        sys.exit(
                          "Wrong parameter settings in TBM_settings.ini loop_time_ms"
                        )
                    main_shunt_millivolt.set_reportFrequency(freqStr)
                    main_shunt_millivolt.registerTimedReportCallback(timedCallback)
                else:
                    this_sensor.set_reportFrequency("OFF")
                    this_sensor.registerTimedReportCallback(None)
        except Exception as exc:
            handle_error(exc)
        this_sensor = this_sensor.nextSensor()
    #
    this_display = YDisplayFirstDisplay()
    while this_display is not None:
        display_arrappend(this_display)
        this_display = this_display.nextDisplay()
#
def deviceArrival(module):
    ## global modules
    try:
        print(modulecondisconformat('', 
                                    module.get_serialNumber(), 
                                    module.get_productName())
        )
    except Exception as exc:
        handle_error(exc)
    modulesappend(module)
    reg_modules()
#
def deviceRemoval(module):
    ## global modules
    try:
        print(modulecondisconformat('dis',
                                    module.get_serialNumber(), 
                                    module.get_productName())
        )
    except Exception as exc:
        handle_error(exc)
    modulesremove(module)
    reg_modules()
#
def remaining_hours(curr_a):
    ## Calculate remaining time in hours
    ret = 0.0
    return ret
#
def effective_cap(curr_a):
    if (curr_a <= -1.0):
        curr_a *= -1
        ec = float( pow((ah_cap_20/(curr_a * 20)), \
                        (peukert_exp - 1)) * ah_cap_20)
        print(ecformat(ec))
        return ec
    else:
        return ah_cap_20
#
def reset_ah():
    global calc_efficiency, cycle_ah_in, cycle_ah_out, netto_ah
    cycle_ah_in = 0
    cycle_ah_out = 0
    netto_ah = 0
    if (b_full_once):
        temp = (ah_cap_20 * eff_calc_ah_ratio)
        if (ah_in > temp < ah_out):
            calc_efficiency = ah_out/ah_in
        save_to_config()
#
def send_data():
    global sendArr
    tmpArr = []
    tmpArrappend = tmpArr.append
    for a in sendArr:
      ## Single DB request version:
        sendStr = "/%s/input/post.json?time=%s&apikey=%s&node=%s&json={%s}"%(
            emoncmspath, str(a[0]), apikey, str(nodeid), commajoin(a[1])
        )
        ## logging.info("Sending: %s"%(sendStr))
        print("Sending to: " + sendStr)
        connrequest("GET", sendStr)
        response = conngetresponse().read()
        ## logging.info(response)
        print(str(response))
        connclose()
        if (response != b'ok'):
            tmpArrappend(a) ## Failed connections get to try again next time.
      ## Multi DB requests version:
        ## responses = []
        ## if (conns):
            ## tempcount = 0
            ## for con in conns:
                ## con.request("GET", 
                    ## "/%s/input/post.json?time=%s&%s=%s&node=%s&json={%s}"%(
                        ## domainpaths[tempcount][1], str(int(t_send_to_db)), 
                        ## domainpaths[tempcount][2], str(nodeid), 
                        ## ",".join(sendArr)
                    ## )
                ## )
                ## responses.append(str(con.getresponse().read()))
                ## tempcount += 1
    del sendArr[:] ## Deletes content of sendArr instead of assigning a new list.
    if tmpArr:
        for a in tmpArr:
            sendArrappend(a)
#
def beaconOnOff(sensArr,deviation):
    ## Turns module beacon on or off.
    bDeviation = False
    if (sensArr and len(sensArr) > 1):
        arr = []
        arrappend = arr.append
        for sensor in sensArr:
            if (sensor.isOnline()):
                try:
                    arrappend(sensor.get_currentValue())
                    continue
                except Exception as exc:
                    handle_error(exc)
            arrappend(0)
        avg = sum(arr)/len(arr)
        for i in range(len(arr)):
            try:
                if (sensArr[i].isOnline()):
                    if (abs(avg-arr[i]) > deviation):
                        sensArr[i].get_module().set_beacon(YModule.BEACON_ON)
                        bDeviation = True
                    else:
                        sensArr[i].get_module().set_beacon(YModule.BEACON_OFF)
            except Exception as exc:
                handle_error(exc)
    return bDeviation
#
def draw_soc(w,h,l0,soc):
    try:
        ## Draws a battery showing current SOC on first display layer of first display.
        l0.drawRect(16,0,w-3,h-1)

        l0.drawRect(((w-2) if (soc == 1.0) else (w-3)), h/2-3, w-1, h/2+2)
        ## Same functionality as commented block below
        ## if (soc == 1.0):
            ## l0.drawRect(w-2,h/2-3,w-1,h/2+2)
        ## else:
            ## l0.drawRect(w-3,h/2-3,w-1,h/2+2)
        
        if (soc >= 0.0):
            l0.drawBar(16,1,soc*76+16,h-2)
            l0.drawText(16,0,l0.ALIGN.TOP_RIGHT,percformat(int(soc*100)))
        else:
            l0.drawText(19,h/2-1,l0.ALIGN.CENTER_LEFT,"SOC unknown")
    except Exception as exc:
        handle_error(exc)
#
def getSensorValue(sensor):
    if (sensor.isOnline()):
        try:
            if (sensor.get_reportFrequency() == "OFF"):
                return colonformat(sensor.get_friendlyName(), 
                    str(sensor.get_currentValue()))
            else:
                print(colonformat(sensor.get_friendlyName(), 
                    str(sensor.get_currentValue())))
        except Exception as exc:
            handle_error(exc)
    return 0
#
def main_loop(appendStrings):
    global soc, t_send_to_db, last_report_time
    last_report_time = t_loopstart
    YAPIUpdateDeviceList()
    strArr = []
    strArrappend = strArr.append
    ############################################################## DEBUG
    strArrappend("netto_ah:" + str(netto_ah))
    strArrappend("ah_in:" + str(ah_in))
    strArrappend("ah_out:" + str(ah_out))
    strArrappend("cycle_ah_in:" + str(cycle_ah_in))
    strArrappend("cycle_ah_out:" + str(cycle_ah_out))
    ############################################################## DEBUG
    if (appendStrings is not None):
        for string in appendStrings:
            strArrappend(string)
    if (sensor_arr):
        for sensor in sensor_arr:
            tmpValue = getSensorValue(sensor)
            if (tmpValue):
                strArrappend(tmpValue)
        #
        if (soc >= 0.0):
            ## Making sure SOC is within range of 0% and 100%
            if (soc > 1.0):
                soc = 1.0
                print("Note: SOC limited to 100 %")
            strArrappend("SOC:%06.2f"%(soc*100))
            if (soc < (low_bat_warn)):
                low_bat_warn_func()
        #
        ## strArr.append("calc_efficiency:%06.2f"%(calc_efficiency*100))
        print("User set efficiency %:   " + str(efficiency*100))
        if (calc_efficiency):
            print("Calculated efficiency %: " + str(calc_efficiency*100))
        print("b_full_once: " + str(b_full_once))
        print("b_full_now:  " + str(b_full_now))
        #
        ## print("t_loopstart: %f\nt_send_to_db:%f"%(t_loopstart, t_send_to_db))
        if (apikey and domain and nodeid):
            sendArrappend([int(t_loopstart),strArr])
            if ((t_loopstart - t_send_to_db) > save_to_db_time_sec):
                t_send_to_db = t_loopstart
                send_data()
        #
        if (display_arr):
            dispNr = 0
            ## print("%s (Ctrl-C to stop)"%(response))
            for disp in display_arr:
                try:
                    if (disp.isOnline()):
                        disp.resetAll()
                        w = disp.get_displayWidth()
                        h = disp.get_displayHeight()
                        l0 = disp.get_displayLayer(0)
                        l0.clear()
                        if (dispNr == 0):
                            draw_soc(w,h,l0,soc)
                        else:
                            if (dispNr <= len(strArr)):
                                sens = YSensorFindSensor(
                                    strArr[dispNr-1].split(":")[0]
                                )
                                sfName = sens.get_logicalName()
                                if (not sfName):
                                    sfName = sens.get_functionId()
                                l0.drawText(
                                    0,
                                    h/2,
                                    l0.ALIGN.CENTER_LEFT,
                                    "%s: %3.2f %s"%(
                                        sfName, 
                                        float(strArr[dispNr-1].split(":")[1]), 
                                        sens.get_unit()
                                    )
                                )
                            else:
                                draw_soc(w,h,l0,soc)
                        dispNr += 1
                except Exception as exc:
                    handle_error(exc)
        #
        bDeviation = beaconOnOff(block_voltage_arr,block_deviation_beacon)
        if (bDeviation):
            print("WARNING: The voltage of at least one battery block is " 
                + "deviating too much from the average!")
        bDeviation = beaconOnOff(cell_voltage_arr,cell_deviation_beacon)
        if (bDeviation):
            print("WARNING: The voltage of at least one battery cell is " 
                + "deviating too much from the average!")
        ## print(response)

        ## for s in strArr:
            ## print(s)
        print("")
    else:
        print("No sensor connected! Retrying... (Ctrl-C to stop)")
#
def get_main_voltage():
    if (main_voltage_volt and main_voltage_volt.isOnline()):
        try:
            voltage_value_v = main_voltage_volt.get_currentValue()
            if (voltage_value_v > over_voltage):
                over_voltage_warn_func()
            elif (voltage_value_v < under_voltage):
                under_voltage_warn_func()
            return voltage_value_v
        except Exception as exc:
            handle_error(exc)
    return "OFFLINE"
#
def main_shunt_millivolt_function(sensorDevice, avg, diff, end):
    global ah_in, ah_out, b_full_now, b_full_once, cycle_ah_in, cycle_ah_out, \
           netto_ah, soc, t_charged, t_loopstart
    strArr2 = []
    strArr2append = strArr2.append
    current_average_a = ((avg * main_shunt_multiplier) + main_shunt_offset)
    ############################################################## DEBUG
    strArr2append(colonformat("Ushunt_mV",str(avg)))
    ############################################################## DEBUG
    try:
        strArr2append(colonformat(sensorDevice.get_friendlyName(), 
                                  str(current_average_a)))
    except Exception as exc:
        handle_error(exc)
    voltage_value_v = get_main_voltage()
    ## if (main_shunt_millivolt and main_shunt_millivolt.isOnline()):
    print(curaverageaformat(current_average_a))
    ## if (abs(current_average_a) < current_threshold_a):
        ## current_average_a = 0.0
    if (voltage_value_v == "OFFLINE"):
        print("SOC unknown")
    else:
        if (0.0 <= current_average_a <= (ah_cap_20 * charged_a_ratio) 
          and voltage_value_v >= charged_v):
            if (t_loopstart > (t_charged + fully_charged_sec)):
                soc = 1.0
                if (b_full_once):
                    ## if (cycle_ah_in and cycle_ah_out):
                    ah_in += cycle_ah_in
                    ah_out += cycle_ah_out
                reset_ah()
                b_full_now = True
        else:
            t_charged = t_loopstart
            if (b_full_now and current_average_a < 0.0):
                b_full_once = True
                b_full_now = False
        if (current_average_a):
            tmp_ah = current_average_a * (diff/3600.0)
            if (current_average_a < 0.0):
                cycle_ah_out -= tmp_ah
                netto_ah += tmp_ah
            else:
                cycle_ah_in += tmp_ah
                netto_ah += (efficiency * tmp_ah)
        print(nettocycleformat(netto_ah, cycle_ah_out, cycle_ah_in))
        if (b_full_once and ah_cap_20):
            effcap = effective_cap(current_average_a)
            soc = ((netto_ah+effcap)/effcap)
            print("SOC %: " + "%6.2f"%(soc*100))
    t_loopstart = end
    main_loop(strArr2)
    ## send_data(str(end), strArr2)
    ## elif (main_voltage_volt and main_voltage_volt.isOnline()):
        ## Code for SOC by measured voltage
        ## print("SOC by voltage not yet implemented!")
#
def timedCallback(sensorDevice,measure):
    try:
        start = measure.get_startTimeUTC()
        end = measure.get_endTimeUTC()
        diff = float(end) - float(start)
        avg = measure.get_averageValue()
        max = measure.get_maxValue()
        min = measure.get_minValue()
        main_shunt_millivolt_function(sensorDevice, avg, diff, end)
        ## send_data(str(start), ["%s:%s"%(sensorDevice.get_friendlyName(), 
                              ## str(avg))])
    except Exception as exc:
        handle_error(exc)
#
## Registering what functions to use when a device connects or disconnects.
YAPI.RegisterDeviceArrivalCallback(deviceArrival)
YAPI.RegisterDeviceRemovalCallback(deviceRemoval)

## Continue the loop until interrupted.
while True:
    t_loopstart = timetime()
    if (t_loopstart - last_report_time > loop_time_ms / 250):
        YAPIUpdateDeviceList()
    if (main_shunt_millivolt is None):
        ## Unless there's a timed callback, manually poll all sensors.
        try:
            main_loop(None)
        except Exception as exc:
            handle_error(exc)
    YAPISleep(int(loop_time_ms-(((timetime()-starttime)*1000)%loop_time_ms)))
#
