// Declaring page global variables
var path;
var feedlist;
var apikey = "apikey=";
var dropdownlists = [];
var searchFor = [];
var barList = [];
var colours = ["#0000FF", "#FF0000"];
var step = 0.1;
var threshold = 0.5;
var labelsStartString = "Block ";
var chartTitle = "Custom Chart Title";
var barChart;
var timerID;
var timerTime = 60*1000;

function fillSearchForArray(){
  searchFor = [];
  $.ajax({ url: path+"feed/list.json?"+apikey, dataType: 'json', async: false, success: function(data) {feedlist = data;} });
  for (z in feedlist) {
    if (contains(feedlist[z]['name'], "UBB")) {
      searchFor.push(feedlist[z]['name']);
    }
  }
  
  // searchFor.push("Some other feed name");
  
  return searchFor;
}

function enableDisableRemBtn(){
  if (dropdownlists.length > 1){
    $('#remfeedbtn').attr('disabled', false);
  }else{
    $('#remfeedbtn').attr('disabled', true);
  }
  onSelectChanged();
}

function draw_options(searchForString) {
  searchForString = (searchForString || undefined);
  // Build options table html
  var options_html = "";
  options_html += "<div id='box-options" + (dropdownlists.length) + "'>";
  options_html += "<div class='input-prepend'><span class='add-on' style='width: 70px; text-align: right;'>Feed</span>";
  options_html += select_feed(searchForString);
  options_html += "</div></div>";
  $('#box-options').append(options_html);
  enableDisableRemBtn();
  $('#feedid'+(dropdownlists.length-1)).change(function(){onSelectChanged();})
}

// Remove feed dropdownlist
function remove_options() {
  //var options_html = $("#box-options").html();
  dropdownlists.pop();
  //$('#box-options').html(options_html.split('<div id="box-options'+ (dropdownlists.length) +'">')[0]); // Don't use, better method using jQuery below
  $('#box-options'+dropdownlists.length).remove();
  enableDisableRemBtn();
}

// Create a drop down select box with a list of feeds.
function select_feed(searchForString) {
  var feedgroups = [];
  $.ajax({ url: path+"feed/list.json?"+apikey, dataType: 'json', async: false, success: function(data) {feedlist = data;} });
  for (z in feedlist) {
    var group = (feedlist[z].tag === null ? "NoGroup" : feedlist[z].tag);
    if (group!="Deleted") {
      if (!feedgroups[group]) feedgroups[group] = [];
      feedgroups[group].push(feedlist[z]);
    }
  }
  var out = "<select id='feedid" + dropdownlists.length + "' class='options' otype='feed'>";
  dropdownlists[dropdownlists.length] = "feedid" + dropdownlists.length; // Adding feed dropdownlist id to page list
  for (z in feedgroups) {
    out += "<optgroup label='" + z + "'>";
    for (p in feedgroups[z]) {
      out += "<option value='" + feedgroups[z][p]['id'] + "' public='" + feedgroups[z][p]['public'] + "' " + (feedgroups[z][p].name == searchForString ? " selected" : "") + ">" + feedgroups[z][p].id + ": " + feedgroups[z][p].name + "</option>";
    }
    out += "</optgroup>";
  }
  out += "</select>";
  return out;
}

// +/- Feed button event functions
function onAddFeedBtnClick(){ // Add new feed at bottom
  draw_options();
}

function onRemFeedBtnClick(){ // Remove bottom most feed
  remove_options();
}

// Adds click events for the two feed buttons
$('#addfeedbtn').click( function(){onAddFeedBtnClick();} );
$('#remfeedbtn').click( function(){onRemFeedBtnClick();} );

function contains(arr, v){
  return ~arr.indexOf(v);
}

// Filters identical numbers and sorts the array
function sortedUniqueIDs(){
  var feedids = [];
  for (var index in dropdownlists){
    x = $('#'+dropdownlists[index]).val();
    if (!contains(feedids, x)) feedids.push(x);
  }
  return feedids.sort();
}

// Get data
function getData(feedids){
  var feedin = [[], []];
  for (var index in feedids){
    feedin[0].push(feedids[index]);
  }
  $.ajax({ url: path+"feed/fetch.json?ids="+feedin[0].join(","), dataType: 'json', async: false, success: function(data) {feedin[1] = data;} });
  return feedin;
}

function onSelectChanged(){
  barList = [];
  var dataArr = getData(sortedUniqueIDs());
  for (var index in dropdownlists){
    var i = dataArr[0].indexOf($('#feedid'+index).val());
    barList.push([labelsStartString+(Number(index)+1),dataArr[1][i]]);
  }
  if (timerID){
    clearTimeout(timerID);
  }
  if (!(barChart == undefined || barChart === undefined)){
    barChart.destroy();
  }
  updateChart();
}

function updateChart(){
  if (barList.length > 1){
    var labels = [], data = [], colors = [];
    $('.toremove').remove();
    var htmlout = "<tr class='toremove'><td></td><td><p>Voltage (V)</p></td><td><p>Deviation from mean</p></td></tr>";
    for (var z in barList){
      labels.push(barList[z][0]);
      data.push(+(barList[z][1]));
    }
    var arrMean = (data.reduce(function(a, b){return a + b;}) / data.length);
    for (var z in data){
      htmlout += "<tr class='toremove'><td>" + labels[z] + "</td><td>" + data[z] + "</td>";
      data[z] -= arrMean;
      htmlout += "<td>" + data[z].toPrecision(2) + "</td></tr>";
      colors.push((Math.abs(data[z])<threshold?colours[0]:colours[1]));
    }
    $('#tbody').append(htmlout);
    barChart = new Chart($('#comparisonChart'), {
      type: 'bar',
      data: {
        labels: labels,
        datasets: [{
          data: data,
          backgroundColor: colors,
          hoverBackgroundColor: colors
        }]
      },
      options: {
        tooltips: {
          enabled: false
        },
        title: {
          display: true,
          text: 'Custom Title'
        },
        legend: {
          display: false
        },
        scales: {
          yAxes: [{
            scaleLabel: {
              display: true,
              labelString: "Voltage difference from mean value"
            },
            ticks:{
              beginAtZero:false,
              stepSize: step,
              suggestedMax: 1,
              suggestedMin: -1,
            }
          }]
        }
      }
    });
    timerID = setTimeout(onSelectChanged,timerTime);
  }
}

// Show/Hide advanced settings function
function showHideBtnClick(){
  $('#hide1').toggle();
  $('#hide2').toggle();
}

for (var i in fillSearchForArray()){
  draw_options(searchFor[i]);
}
if (dropdownlists.length == 0) draw_options();

// Show/Hide Button Click event
$('#showhidebtn').click( function(){showHideBtnClick();} );
